
import os
import argparse
from urllib.parse import quote # encode special characters
import json
import requests
import pytz
from timezonefinder import TimezoneFinder
from datetime import datetime

WRTFORMT = "%Y:%m:%d %H:%M:%S%z"
# Where to reach IMWS or the embedded IMWS in IMatch
HOST_URL = 'http://127.0.0.1:50519'
REQUEST_TIMEOUT = 10                    # Request timeout in seconds

# This stores the IMWS authentication token after authenticate() has been
# called
auth_token = ''


def access_imatch(endpoint, params, post=True):
    global auth_token
    req = None
    try:
        if post:
            req = requests.post(HOST_URL + endpoint, params=params,
                                timeout=REQUEST_TIMEOUT)
        else:
            req = requests.get(HOST_URL + endpoint, params=params,
                               timeout=REQUEST_TIMEOUT)

        if req.status_code == requests.codes.ok:
            if req.text == "":
                return True, ""
            # Convert the JSON response
            response = json.loads(req.text)
            return True, response
        elif req.status_code == requests.codes.no_content:
            return True, ""
        else:
            # Raise the exception matching the HTTP status code
            req.raise_for_status()

        return False, ""

    except requests.exceptions.RequestException as re:
        vprint(re)
        return False, ""

    except Exception as ex:
        vprint(endpoint, ex)
        return False, ""


def authenticate(user=None):
    """ Authenticate against IMWS and set the auth_token variable
        to the returned authentication token. We need this for all other endpoints. """

    global auth_token
    if user is None:
        user = os.getlogin()
    params = {
        'id': user,
        'password': '',
        'appid': ''}
    ans, response = access_imatch('/v1/authenticate', params)

    if ans:
        auth_token = response["auth_token"]
        return True
    else:
        print("Authentication failed")
    return False


def doUpdate(taskList):
    if len(taskList) > 0:
        # If too many changes, request fails.  So do in batches.
        chunks = [taskList[x:x+40] for x in range(0, len(taskList), 40)]
        for c in chunks:
            tasks = json.dumps(c, separators=(',', ':'))
            params = {
                'auth_token': auth_token,
                'tasks': tasks
            }
            ans, response = access_imatch('/v1/metadata', params)
            if ans:
                vprint("  {0} Changes made".format(len(c)))
            else:
                print("  Failed to make {0} changes".format(len(c)))
    else:
        vprint("  No changes to make")


def process_selected_files(deftz):
    params = {
        'auth_token': auth_token,
        'idlist': '@imatch.filewindow.active.selection',
        'fields': 'id,filename',
        'varDTO': '{File.MD.Exif::Main\\36867\\DateTimeOriginal\\0|value:rawfrm}',
        'varXDTO': '{File.MD.XMP::exif\\DateTimeOriginal\\DateTimeOriginal\\0|value:rawfrm}',
        'varXCD': '{File.MD.XMP::xmp\\CreateDate\\CreateDate\\0|value:rawfrm}',
        'varDSC': '{File.MD.XMP::photoshop\\DateCreated\\DateCreated\\0|value:rawfrm}',
        'varLat': '{File.MD.gpslatitude|value:rawfrm}',
        'varLng': '{File.MD.gpslongitude|value:rawfrm}'
    }
    ans, response = access_imatch('/v1/files', params, post=False)
    if ans:
        vprint("Number of files: ", len(response['files']))
        tasks = get_tasks(response['files'], deftz)
        doUpdate(tasks)
    else:
        print("Failed to get metadata")
    return


def vprint(*txt):
    if args.v:
        print(*txt)


def get_tasks(fls, deftz):
    tf = TimezoneFinder()
    tasks = []
    for f in fls:
        update = False
        if f["DTO"] != "": # base new date on Exif:DateTimeOriginal
            update = True
            natdt = f["DTO"][:19].replace("/", ":")
        elif f["XDTO"] != "": # base new date on XMP:DateTimeOriginal
            update = True
            natdt = f["XDTO"][:19].replace("/", ":")
        elif f["DSC"] != "": # base new date on Date Subject Created
            update = True
            natdt = f["DSC"][:19].replace("/", ":")
        if update:
            dto = datetime.strptime(natdt, "%Y:%m:%d %H:%M:%S")
            tz = get_tz(tf, f, deftz)
            vprint(f['fileName'], " - timezone - ", tz)
            dtotz = pytz.timezone(tz).localize(dto)
            vprint("  Date Subject Created: ", dto,
                   " - Localized to: ", dtotz.strftime(WRTFORMT))
            newdate = dtotz.strftime(WRTFORMT)
            newdate = quote(newdate[:-2] + ":" + newdate[-2:])
            # populate Date Sybject Created
            task = {"id": f["id"],
                    "tag": "XMP::photoshop\\DateCreated\\DateCreated",
                    "op": "set",
                    "value": newdate}
            tasks.append(task)
            if args.createdate: # Populate Create Date
                task = {"id": f["id"],
                        "tag": "XMP::xmp\\CreateDate\\CreateDate",
                        "op": "set",
                        "value": newdate}
            else:
                '''
                Don't change Create Date.  
                BUT writing back a changed Date Subject Created results in
                other mapped fields changing.  In some cases this can propagate to 
                cause changes in Create Date.  To avoid this, re-populate Create Date
                with its current value.
                '''
                task = {"id": f["id"],
                    "tag": "XMP::xmp\\CreateDate\\CreateDate",
                    "op": "set",
                    "value": f["XCD"]}
            tasks.append(task)
    return tasks


def get_tz(tf, tag, deftz):
    timez = deftz  # default assumption
    if (len(tag['Lng']) > 0) and (len(tag['Lat']) > 0):
        timez = tf.certain_timezone_at(
            lat=float(tag["Lat"]),
            lng=float(tag["Lng"]),
        )
    return timez


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-deftz",
        help="Specify default timezone, (e.g. -deftz Europe/Oslo) for use when there are no GPS coords. If omitted, 'Europe/London' is default.",
        default='Europe/London',
        required=False)
    parser.add_argument(
        "-user",
        help="Specify user name alias if this has been set in iMatch Application settings.  Otherwise not needed",
        required=False)
    parser.add_argument(
        '-createdate',
        help='Normally Date Subject Created is populated.  By specifying this option, the same contents are also written to Create Date',
        action='store_true',
        required=False)
    parser.add_argument(
        '-v',
        help='Verbose output',
        action='store_true',
        required=False)
    parser.add_argument(
        '-np',
        help='Do not pause output for confirmations',
        action='store_true',
        required=False)
    args = parser.parse_args()
    ans = ""
    if not args.np:
        ans = input("Warning - this cannot be undone! Proceed(Y/n)?")
    if (ans.upper() == "Y") or args.np:
        if authenticate(args.user):
            if args.deftz is not None:
                if args.deftz in pytz.all_timezones:
                    process_selected_files(deftz=args.deftz)
                else:
                    print("Invalid timezone")
            else:
                process_selected_files()
        if (not args.np) or (args.v):
            input("Press return to continue")
